import pytest
import threading

from zaber.serial import AsciiDevice, AsciiCommand

@pytest.fixture
def device(asciiserial):
    return AsciiDevice(asciiserial, 1)

def test_constructor(asciiserial):
    ad = AsciiDevice(asciiserial, 1)
    assert(ad.port == asciiserial)
    assert(ad.address == 1)

def test_send_command(device, fake):
    t = threading.Thread(target = device.send, args = (AsciiCommand("home"), ))
    t.start()
    fake.expect("/1 0 home\r\n")
    fake.send("@01 0 OK BUSY WR 0\r\n")
    t.join()

def test_send_command_with_device_number(device, fake):
    t = threading.Thread(target = device.send, args = (AsciiCommand("1 home"), ))
    t.start()
    fake.expect("/1 0 home\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    t.join()

def test_send_str(device, fake):
    t = threading.Thread(target = device.send, args = ("home", ))
    t.start()
    fake.expect("/1 0 home\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    t.join()

def test_send_str_with_device_number(device, fake):
    t = threading.Thread(target = device.send, args = ("1 move vel 1000", ))
    t.start()
    fake.expect("/1 0 move vel 1000\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    t.join()

def test_send_str_with_axis_number(device, fake):
    t = threading.Thread(target = device.send, args = ("1 1 move rel -1000", ))
    t.start()
    fake.expect("/1 1 move rel -1000\r\n")
    fake.send("@01 1 OK BUSY -- 0\r\n")
    t.join()

def test_send_str_with_multiple_arguments(device, fake):
    t = threading.Thread(target = device.send, args=("/tools setcomm 9600 1\r\n", ))
    t.start()
    fake.expect("/1 0 tools setcomm 9600 1\r\n")
    fake.send("@01 0 OK IDLE NU 0")
    t.join()

def test_home(device, fake):
    t = threading.Thread(target = device.home)
    t.start()
    fake.expect("/1 0 home\r\n")
    fake.send("@01 0 OK BUSY WR 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK BUSY WR 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK IDLE -- 0\r\n")
    t.join()

def test_move_abs(device, fake):
    t = threading.Thread(target = device.move_abs, args = (1000, ))
    t.start()
    fake.expect("/1 0 move abs 1000\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK IDLE -- 0\r\n")
    t.join()

def test_move_rel(device, fake):
    t = threading.Thread(target = device.move_rel, args = (1322, ))
    t.start()
    fake.expect("/1 0 move rel 1322\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK IDLE -- 0\r\n")
    t.join()

def test_move_vel(device, fake):
    t = threading.Thread(target = device.move_vel, args = (3243, ))
    t.start()
    fake.expect("/1 0 move vel 3243\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    t.join()

def test_stop(device, fake):
    t = threading.Thread(target = device.stop)
    t.start()
    fake.expect("/1 0 stop\r\n")
    fake.send("@01 0 OK BUSY NI 0\r\n")
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK IDLE NI 0\r\n")
    t.join()

def test_get_status(device, fake):
    t = threading.Thread(target = device.get_status)
    t.start()
    fake.expect("/1 0\r\n")
    fake.send("@01 0 OK BUSY -- 0\r\n")
    t.join()



from zaber.serial import AsciiCommand

def test_string_constructor():
    ac = AsciiCommand("home")
    assert(ac.device_address == 0)
    assert(ac.axis_number == 0)
    assert(ac.data == "home")

def test_leading_slash():
    ac = AsciiCommand("/move rel 1000")
    assert(ac.device_address == 0)
    assert(ac.axis_number == 0)
    assert(ac.data == "move rel 1000")

def test_trailing_return_newline():
    ac = AsciiCommand("tools setcomm 9600 1\r\n")
    assert(ac.device_address == 0)
    assert(ac.axis_number == 0)
    assert(ac.data == "tools setcomm 9600 1")

def test_device_address():
    ac = AsciiCommand("/1 home")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 0)
    assert(ac.data == "home")

def test_axis_number():
    ac = AsciiCommand("/12 3 move abs 1000\r\n")
    assert(ac.device_address == 12)
    assert(ac.axis_number == 3)
    assert(ac.data == "move abs 1000")

def test_message_id_present():
    ac = AsciiCommand("1 2 3 home")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 2)
    assert(ac.message_id == 3)
    assert(ac.data == "home")

def test_message_id_is_none():
    ac = AsciiCommand("1 1 home")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 1)
    assert(ac.message_id == None)
    assert(ac.data == "home")

def test_multiple_arguments():
    ac = AsciiCommand(1, 2, "home")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 2)
    assert(ac.data == "home")

def test_arguments_without_axis_number():
    ac = AsciiCommand(3, "asdf")
    assert(ac.device_address == 3)
    assert(ac.axis_number == 0)
    assert(ac.data == "asdf")

def test_arguments_with_message_id():
    ac = AsciiCommand(5, 6, 7, "home")
    assert(ac.device_address == 5)
    assert(ac.axis_number == 6)
    assert(ac.message_id == 7)
    assert(ac.data == "home")

def test_empty_constructor():
    ac = AsciiCommand()
    assert(ac.device_address == 0)
    assert(ac.axis_number == 0)
    assert(ac.data == "")

def test_multiple_string_arguments():
    ac = AsciiCommand("1", "2", "4", "move rel", "10000")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 2)
    assert(ac.message_id == 4)
    assert(ac.data == "move rel 10000")

def test_multiple_string_arguments_with_integer_arguments():
    ac = AsciiCommand(1, 2, "move abs", "4000")
    assert(ac.device_address == 1)
    assert(ac.axis_number == 2)
    assert(ac.data == "move abs 4000")

def test_command_parameter_as_integer():
    ac = AsciiCommand(1, "move rel", 5000)
    assert(ac.device_address == 1)
    assert(ac.data == "move rel 5000")

def test_binary_strings():
    ac = AsciiCommand(1, b"move abs", 4000)
    assert(ac.device_address == 1)
    assert(ac.data == "move abs 4000")
    ac2 = AsciiCommand(b"/1 0 tools setcomm 9600 1\r\n")
    assert(ac2.device_address == 1)
    assert(ac2.axis_number == 0)
    assert(ac2.data == "tools setcomm 9600 1")

def test_print():
    command_string = "/1 0 get system.temperature\r\n"
    ac = AsciiCommand(command_string)
    printed_string = str(ac)
    assert(printed_string == command_string.rstrip("\r\n"))



"""The 'ascii' module contains all classes related to the ASCII
protocol.
"""

import serial
import time
import logging

from .exceptions import TimeoutError, UnexpectedReplyError

# See https://docs.python.org/2/howto/logging.html#configuring-logging-
# for-a-library for info on why we have these two lines here.
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class AsciiAxis(object):
    """Represents one axis of an ASCII device.
    
    Attributes:
        parent: An AsciiDevice which represents the device which has
            this axis.
        number: The number of this axis. 1-9.
    """

    def __init__(self, device, number):
        """
        Args:
            device: An AsciiDevice which is the parent of this axis.
            number: The number of this axis. Must be 1-9.

        Raises:
            ValueError: The axis number was not between 1 and 9.
        """
        if number < 1 or number > 9:
            raise ValueError("Axis number must be between 1 and 9.")
        self.number = number
        self.parent = device

    def send(self, message):
        """Sends a message to the axis and then waits for a reply.

        Args:
            message: A string or AsciiCommand object containing a
                command to be sent to this axis.

        Notes:
            Regardless of the device address or axis number supplied in
            (or omitted from) the message passed to this function, this
            function will always send the command to only this axis.
            
            Though this is intended to make sending commands to a
            particular axis easier by allowing the user to pass in a
            "global command" (ie. one whose target device and axis are
            both 0), this can result in some unexpected behaviour. For
            example, if the user tries to call send() with an
            AsciiCommand which has a different target axis number than
            the number of this axis, they may be surprised to find that
            the command was sent to this axis rather than the one
            originally specified in the AsciiCommand.

        Examples:
            Since send() will automatically set (or overwrite) the
            target axis and device address of the message, all of the
            following calls to send() will result in identical ASCII
            messages being sent to the serial port::

                >>> axis.send("home")
                >>> axis.send(AsciiCommand("home"))
                >>> axis.send("0 0 home")
                >>> axis.send("4 8 home")
                >>> axis.send(AsciiCommand(1, 4, "home"))

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the reply received.
        """
        if isinstance(message, (str, bytes)):
            message = AsciiCommand(message)

        # Always send the AsciiCommand to *this* axis.
        message.axis_number = self.number

        reply = self.parent.send(message)
        if reply.axis_number != self.number:
            raise UnexpectedReplyError("Received a reply from an "
                    "unexpected axis: axis {}".format(reply.axis_number),
                    reply)
        return reply

    def home(self):
        """Sends the "home" command, then polls the axis until it is
        idle.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the first reply
            received.
        """
        reply = self.send("home")
        self.poll_until_idle()
        return reply

    def move_abs(self, position, blocking = True):
        """Sends the "move abs" command to the axis to move it to the
        specified position, then polls the axis until it is idle.

        Args:
            position: An integer representing the position in
                microsteps to which to move the axis.
            blocking: An optional boolean, True by default. If set to
                False, this function will return immediately after
                receiving a reply from the device, and it will not poll
                the device further.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the first reply
            received.
        """
        reply = self.send("move abs {0:d}".format(position))
        if blocking: self.poll_until_idle()
        return reply

    def move_rel(self, distance, blocking = True):
        """Sends the "move rel" command to the axis to move it by the
        specified distance, then polls the axis until it is idle.
        
        Args:
            distance: An integer representing the number of microsteps
                by which to move the axis.
            blocking: An optional boolean, True by default. If set to
                False, this function will return immediately after
                receiving a reply from the device, and it will not poll
                the device further.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the first reply
            received.
        """
        reply = self.send("move rel {0:d}".format(distance))
        if blocking: self.poll_until_idle()
        return reply

    def move_vel(self, speed, blocking = False):
        """Sends the "move vel" command to make the axis move at the
        specified speed.

        Args:
            speed: An integer representing the speed at which to move
                the axis.
            blocking: An optional boolean, False by default. If set to
                True, this function will poll the device repeatedly
                until it reports that the axis is idle.

        Notes:
            Unlike the other two move commands, move_vel() does not by
            default poll the axis until it is idle. move_vel() will
            return immediately after receiving a response from the
            device unless the "blocking" argument is set to True.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the first reply
            received.
        """
        reply = self.send("move vel {0:d}".format(speed))
        if blocking: self.poll_until_idle()
        return reply

    def stop(self):
        """Sends the "stop" command to the axis.

        Notes:
            The stop command can be used to pre-empt any movement
            command in order to stop the axis early. 

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the first reply
            received.
        """
        reply = self.send("stop")
        self.poll_until_idle()
        return reply

    def get_status(self):
        """Queries the axis for its status and returns the result.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.
        
        Returns:
            A string containing either "BUSY" or "IDLE", depending on
            the response received from the axis.
        """
        return self.send("").device_status

    def poll_until_idle(self):
        """Polls the axis and blocks until the device reports that the
        axis is idle.

        Raises:
            UnexpectedReplyError: The reply received was not sent by the
                expected device and axis.

        Returns: An AsciiReply object containing the last reply
            received.
        """
        return self.parent.poll_until_idle(self.number)

class AsciiCommand(object):
    """Models a single command in Zaber's ASCII protocol.

    Attributes:
        device_address: An integer representing the address of the 
            device to which to send this command.
        axis_number: The integer number of the particular axis which 
            should execute this command. An axis number of 0 specifies
            that all axes should execute the command, or that the
            command is "device scope".
        message_id: Optional. An integer to be used as a message ID.
            If a command has a message ID, then the device will send a
            reply with a matching message ID. A message_id value of
            None indicates that a message ID is not to be used.
            0 is a valid message ID.
        data: The bulk of the command. data includes a valid ASCII
            command and any parameters of that command, separated by
            spaces. A data value of "" (the empty string) is valid,
            and is often used as a "get status" command to query
            whether a device is busy or idle.
    """

    def __init__(self, *args):
        r"""
        Args:
            *args: A number of arguments beginning with 0 to 3 integers
                followed by one or more strings. 
                
        Notes:
            For every absent integer argument to ``__init__``, any string 
            argument(s) will be examined for leading integers. The first
            integer found (as an argument or as the leading part of a
            string) will set the ``device_address`` property, the second
            integer will be taken as the ``axis_number`` property, and
            the third integer found will be the ``message_id`` property.

            When a string argument contains text which can not be
            interpreted as an integer, all arguments which follow it
            are considered to be a part of the data. This is consistent
            with how ASCII commands are parsed by the Zaber device
            firmware.

            All leading '/' and trailing '\\r\\n' characters in string
            arguments are stripped when the arguments are parsed.

        Examples:
            The flexible argument structure of this constructor allows 
            commands to be constructed by passing in integers followed 
            by a command and its parameters, or by passing in one
            fully-formed, valid ASCII command string, or a mix of the
            two if the user desires.

            For example, all of the following constructors will create
            identical AsciiCommand objects::

                >>> AsciiCommand("/1 0 move abs 10000\r\n")
                >>> AsciiCommand("1 move abs 10000")
                >>> AsciiCommand(1, 0, "move abs 10000")
                >>> AsciiCommand(1, "move abs 10000")
                >>> AsciiCommand("1", "move abs", "10000")
                >>> AsciiCommand(1, "move abs", 10000)

        Raises:
            TypeError: An argument was passed to the constructor which
                was neither an integer nor a string.
        """
        self.data = ''
        attributes = iter(["device_address", "axis_number", "message_id"])
        for arg in args:
            if isinstance(arg, int):
                try: 
                    # If self.data has got something in it,
                    # then all remaining arguments are also data.
                    if self.data: raise StopIteration
                    next_attr = next(attributes)
                    setattr(self, next_attr, arg)
                except StopIteration:
                    self.data = ' '.join([self.data, str(arg)]) if self.data \
                            else str(arg)

            elif isinstance(arg, (bytes, str)):
                if isinstance(arg, bytes):
                    arg = arg.decode()

                # Trim leading '/' and trailing "\r\n".
                arg = arg.lstrip('/')
                arg = arg.rstrip('\r\n')

                tokens = arg.split(' ')
                for i, token in enumerate(tokens):
                    try: 
                        # As above: if data has already been found,
                        # all remaining arguments/tokens are also data.
                        if self.data: raise StopIteration
                        num = int(token) # Is it a number?
                        next_attr = next(attributes) # If it *is* a number...
                        setattr(self, next_attr, num) # ...set the next attribute.
                    except (ValueError, StopIteration):  
                        # If token is not a number, or if we are out of 
                        # attributes, the remaining text is data.
                        data = ' '.join(tokens[i:])
                        self.data = ' '.join([self.data, data]) if self.data \
                                else data
                        break
            else:
                raise TypeError("All arguments to AsciiCommand() must be "
                        "either strings or integers. An argument of type "
                        "{0:s} was passed.".format(str(type(arg))))

        # Set remaining attributes.
        if not hasattr(self, "device_address"): self.device_address = 0 
        if not hasattr(self, "axis_number"): self.axis_number = 0
        if not hasattr(self, "message_id"): self.message_id = None

    def encode(self):
        """Return a valid ASCII command based on this object's 
        attributes.
        
        The string returned by this function is a fully valid command,
        formatted according to Zaber's `Ascii Protocol Manual`_. 

        Returns:
            A valid, fully-formed ASCII command.
        """
        if self.message_id is not None:
            if self.data:
                return "/{0:d} {1:d} {2:d} {3:s}\r\n".format(
                        self.device_address, 
                        self.axis_number,
                        self.message_id, 
                        self.data).encode()
            else: return "/{0:d} {1:d} {2:d}\r\n".format(
                    self.device_address,
                    self.axis_number, 
                    self.message_id).encode()

        if self.data:
            return "/{0:d} {1:d} {2:s}\r\n".format(
                    self.device_address,
                    self.axis_number, 
                    self.data).encode()
        else: return "/{0:d} {1:d}\r\n".format(
                self.device_address,
                self.axis_number).encode()

    def __str__(self):
        """Returns an encoded ASCII command, without the newline
        terminator.

        Returns:
            A string containing an otherwise-valid ASCII command,
            without the newline (ie. "\r\n") at the end of the string
            for ease of printing.
        """
        string = self.encode().rstrip(b"\r\n")
        # A little bit of type-checking for Python 2/3 compatibility.
        if not isinstance(string, str):
            string = string.decode()
        return string

class AsciiDevice(object):
    """Represents an ASCII device.
    
    Attributes:
        port: The port to which this device is connected.
        address: The address of this device. 1-99.
    """

    def __init__(self, port, address):
        """
        Args:
            port: An AsciiSerial object representing the port to which
                this device is connected. 
            address: An integer representing the address of this
                device. It must be between 1-99.

        Raises:
            ValueError: The address was not between 1 and 99.
        """
        if address < 1 or address > 99:
            raise ValueError("Address must be between 1 and 99.")
        self.address = address
        self.port = port

    def axis(self, number):
        """Returns an AsciiAxis with this device as a parent and the
        number specified.

        Args:
            number: The number of the axis. 1-9.

        Notes:
            This function will always return a *new* AsciiAxis instance.
            If you are working extensively with axes, you may want to
            create just one set of AsciiAxis objects by directly using
            the AsciiAxis constructor instead of this function to avoid
            creating lots and lots of objects.

        Returns:
            A new AsciiAxis instance to represent the axis specified.
        """
        return AsciiAxis(self, number)

    def send(self, message):
        r"""Sends a message to the device, then waits for a reply.

        Args:
            message: A string or AsciiCommand representing the message
                to be sent to the device.

        Notes:
            Regardless of the device address specified in the message,
            this function will always send the message to this device.
            The axis number will be preserved. 

            This behaviour is intended to prevent the user from
            accidentally sending a message to all devices instead of
            just one. For example, if ``device1`` is an AsciiDevice 
            with an address of 1, device1.send("home") will send the
            ASCII string "/1 0 home\\r\\n", instead of sending the
            command "globally" with "/0 0 home\\r\\n".

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the reply received.
        """
        if isinstance(message, (str, bytes)):
           message = AsciiCommand(message) 

        # Always send an AsciiCommand to *this* device.
        message.device_address = self.address

        self.port.write(message)

        reply = self.port.read()
        if (reply.device_address != self.address
            or reply.axis_number != message.axis_number
            or reply.message_id != message.message_id):
            raise UnexpectedReplyError("Received an unexpected reply from "
                    "device with address {0:d}, axis {1:d}".format(
                        reply.device_address, reply.axis_number),
                    reply)
        return reply

    def poll_until_idle(self, axis_number = 0):
        """Polls the device's status, blocking until it is idle.

        Args:
            axis_number: An optional integer specifying a particular
                axis whose status to poll. axis_number must be between
                0 and 9. If provided, the device will only report the
                busy status of the axis specified. When omitted, the
                device will report itself as busy if any axis is moving.

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the last reply received.
        """
        while True:
            reply = self.send(AsciiCommand(self.address, axis_number, ""))
            if reply.device_status == "IDLE": break
            time.sleep(0.05)
        return reply

    def home(self):
        """Sends the "home" command, then polls the device until it is
        idle.

        Returns:
            An AsciiReply containing the first reply received.
        """
        reply = self.send("home")
        self.poll_until_idle()
        return reply

    def move_abs(self, position, blocking = True):
        """Sends the "move abs" command to the device to move it to the
        specified position, then polls the device until it is idle.

        Args:
            position: An integer representing the position in
                microsteps to which to move the device.
            blocking: An optional boolean, True by default. If set to
                False, this funciton will return immediately after
                receiving a reply from the device and it will not poll
                the device further.

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the first reply received.
        """
        reply = self.send("move abs {0:d}".format(position))
        if blocking: self.poll_until_idle()
        return reply

    def move_rel(self, distance, blocking = True):
        """Sends the "move rel" command to the device to move it by the
        specified distance, then polls the device until it is idle.
        
        Args:
            distance: An integer representing the number of microsteps
                by which to move the device.
            blocking: An optional boolean, True by default. If set to
                False, this function will return immediately after
                receiving a reply from the device, and it will not poll
                the device further.

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the first reply received.
        """
        reply = self.send("move rel {0:d}".format(distance))
        if blocking: self.poll_until_idle()
        return reply

    def move_vel(self, speed, blocking = False):
        """Sends the "move vel" command to make the device move at the
        specified speed.

        Args:
            speed: An integer representing the speed at which to move
                the device.
            blocking: An optional boolean, False by default. If set to
                True, this function will poll the device repeatedly
                until it reports that it is idle.

        Notes:
            Unlike the other two move commands, move_vel() does not by
            default poll the device until it is idle. move_vel() will
            return immediately after receiving a response from the
            device unless the "blocking" argument is set to True.

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the first reply received.
        """
        reply = self.send("move vel {0:d}".format(speed))
        if blocking: self.poll_until_idle()
        return reply

    def stop(self):
        """Sends the "stop" command to the device.

        Notes:
            The stop command can be used to pre-empt any movement
            command in order to stop the device early. 

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            An AsciiReply containing the first reply received.
        """
        reply = self.send("stop")
        self.poll_until_idle()
        return reply

    def get_status(self):
        """Queries the device for its status and returns the result.

        Raises:
            UnexpectedReplyError: The reply received was not sent by
                the expected device.

        Returns:
            A string containing either "BUSY" or "IDLE", depending on
            the response received from the device.
        """
        return self.send("").device_status

class AsciiReply(object):
    """Models a single reply in Zaber's ASCII protocol.

    Attributes:
        message_type: A string of length 1 containing either '@', '!',
            or '#', depending on whether the message type was a 
            "reply", "alert", or "info", respectively. Most messages
            received from Zaber devices are of type "reply", or '@'.
        device_address: An integer between 1 and 99 representing the
            address of the device from which the reply was sent.
        axis_number: An integer between 0 and 9 representing the axis
            from which the reply was sent. An axis number of 0
            represents a reply received from the device as a whole.
        message_id: An integer between 0 and 255 if present, or None
            otherwise.
        reply_flag: A string of length two, containing either "OK" or
            "RJ", depending on whether the command was accepted or
            rejected by the device. Value will be None for device replies
            that do not have a reply flag, such as info and alert messages.
        device_status: A string of length 4, containing either "BUSY"
            or "IDLE", depending on whether the device is moving or 
            stationary.
        warning_flag: A string of length 2, usually "--". If it is not
            "--", it will be one of the two-letter warning flags
            described in the `Warning Flags section`_ of the Ascii
            Protocol Manual.
        data: A string containing the response data.
        checksum: A string of length 2 containing two characteres
            representing a hexadecimal checksum, or None if a checksum
            was not found in the reply.

    .. _Warning Flags section: http://www.zaber.com/wiki/Manuals/ASCII_
        Protocol_Manual#Warning_Flags
    """

    def __init__(self, reply_string):
        """
        Args:
            reply_string: A string in one of the formats described in
                Zaber's `Ascii Protocol Manual`_. It will be parsed by
                this constructor in order to populate the attributes of
                the new AsciiReply.

        Raises:
            ValueError: The string could not be parsed.

        .. _Ascii Protocol Manual: http://www.zaber.com/wiki/Manuals/AS
            CII_Protocol_Manual
        """
        reply_string = reply_string.strip("\r\n")

        if len(reply_string) < 5:
            raise ValueError("Reply string too short to be a valid reply.")

        # CHECK CHECKSUM
        # All message types could have a checksum.
        if reply_string[-3] == ':':
            self.checksum = reply_string[-2:]
            reply_string = reply_string[:-3]
            # Test checksum
            sum = 0
            for ch in reply_string[1:]:
                try: sum += ord(ch)
                except TypeError: sum += ch     # bytes() elements are ints.
            # Truncate to last byte and XOR + 1, as per the LRC.
            # Convert to HEX but keep only last 2 digits, left padded by 0's
            correct_checksum = "{:02X}".format(((sum & 0xFF) ^ 0xFF) + 1)[-2:]
            if self.checksum != correct_checksum:
                raise ValueError("Checksum incorrect. Found {:s}, expected "
                        "{:s}. Possible data corruption detected.".format(
                            self.checksum, correct_checksum))
        else: 
            self.checksum = None

        # SET ATTRIBUTES
        self.message_type = reply_string[0]

        # We iterate over the attributes so we can use next()
        # instead of hardcoding string indices.
        tokens = iter(reply_string[1:].split())

        try:
            # All replies have a device address & axis number.
            self.device_address = int(next(tokens))
            self.axis_number = int(next(tokens))

            # @ is the "Reply" type
            if reply_string[0] == '@':
                t = next(tokens)
                try:
                    self.message_id = int(t)
                    self.reply_flag = next(tokens)
                except ValueError:
                    self.message_id = None
                    self.reply_flag = t

                self.device_status = next(tokens)
                self.warning_flag = next(tokens)
                self.data = next(tokens)

            # # is the "Info" type
            elif reply_string[0] == '#':
                self.device_status = None
                self.warning_flag = None
                self.reply_flag = None
                try:
                    t = next(tokens)
                    try:
                        self.message_id = int(t)
                        self.data = next(tokens)
                    except StopIteration:
                        self.data = ''
                    except ValueError:
                        self.message_id = None
                        self.data = t
                except StopIteration:
                    self.message_id = None
                    self.data = ''

            # ! is the "Alert" type
            elif reply_string[0] == '!':
                self.message_id = None
                self.device_status = next(tokens)
                self.warning_flag = next(tokens)
                self.data = None
                self.reply_flag = None
                return  # Return early to leave the data field as None

            else:
                raise ValueError("Invalid response type: %c" % (reply_string[0],))

        except StopIteration:
            raise ValueError("Incomplete response: {}".format(reply_string))

        # Add any remaining tokens together as the data field.
        try:
            while True:
                self.data = self.data + ' ' + next(tokens)
        except StopIteration:
            pass

    def encode(self):
        """Encodes the AsciiReply's attributes back into a valid string
        resembling the string which would have created the AsciiReply.

        Returns:
            A string in the format described in Zaber's `Ascii Protocol
            Manual`_.

        .. _Ascii Protocol Manual: http://www.zaber.com/wiki/Manuals/AS
            CII_Protocol_Manual
        """
        if self.message_type == '@':
            retstr = "@{:02d} {:d} {:s} {:s} {:s} {:s}".format(
                    self.device_address, self.axis_number, self.reply_flag,
                    self.device_status, self.warning_flag, self.data) \
                    if self.message_id is None else \
                    "@{:02d} {:d} {:02d} {:s} {:s} {:s} {:s}".format(
                            self.device_address, self.axis_number,
                            self.message_id, self.reply_flag,
                            self.device_status, self.warning_flag, self.data)
        elif self.message_type == '#':
            retstr = "#{:02d} {:d} {:s}".format(self.device_address,
                    self.axis_number, self.data) \
                    if self.message_id is None else \
                    "#{:02d} {:d} {:02d} {:s}".format(
                            self.device_address, self.axis_number,
                            self.message_id, self.data)
        elif self.message_type == '!':
            retstr = "!{:02d} {:d} {:s} {:s}".format(self.device_address,
                    self.axis_number, self.device_status, self.warning_flag) \
                    if self.message_id is None else \
                    "!{:02d} {:d} {:s} {:s}".format(self.device_address,
                            self.axis_number, self.message_id, 
                            self.device_status, self.warning_flag)

        if self.checksum is not None:
            return "{:s}:{:s}\r\n".format(retstr, self.checksum)
        else:
            return "{:s}\r\n".format(retstr)

    def __str__(self):
        """Returns a reply string resembling the string which would have
        created this AsciiReply.

        Returns:
            The same string as is returned by encode().
        """
        return self.encode()

class AsciiSerial(object):
    """A class for interacting with Zaber devices using the ASCII protocol.

    Attributes:
        baudrate: An integer representing the desired communication
            baud rate. Valid bauds are 115200, 57600, 38400, 19200, and
            9600.
        timeout: A number representing the number of seconds to wait
            for input before timing out. Floating-point numbers can be
            used to specify times shorter than one second. A value of
            None can also be used to specify an infinite timeout. A
            value of 0 specifies that all reads and writes should be 
            non-blocking (return immediately without waiting). Defaults
            to 5.
    """
    def __init__(self, port, baud = 115200, timeout = 5, inter_char_timeout = 0.01):
        """
        Args:
            port: A string containing the name or URL of the serial port to
                which to connect.
            baud: An integer representing the baud rate at which to
                communicate over the serial port.
            timeout: A number representing the number of seconds to wait
                for a reply. Fractional numbers are accepted and can be
                used to specify times shorter than a second.
            inter_char_timeout : A number representing the number of seconds
                to wait between bytes in a reply. If your computer is bad at
                reading incoming serial data in a timely fashion, try 
                increasing this value.

        Notes: 
            When *port* is not None, this constructor immediately
            opens the serial port. There is no need to call open()
            after creating this object, unless you passed None as
            *port*.

        Raises:
            ValueError: An invalid baud rate was specified.
        """
        if not isinstance(port, str):
            raise TypeError("port must be a string.")
        try:
            self._ser = serial.serial_for_url(port, do_not_open=True)
            self._ser.baudrate = baud
            self._ser.timeout = timeout
            self._ser.interCharTimeout = inter_char_timeout
            self._ser.open()
        except AttributeError:
            # serial_for_url not supported; use fallback
            self._ser = serial.Serial(port, baud, timeout = timeout, interCharTimeout=inter_char_timeout)

    def write(self, command):
        """Writes a command to the serial port.

        Args:
            command: A string or AsciiCommand representing a command
                to be sent.
        """
        if isinstance(command, (str, bytes)):
            command = AsciiCommand(command)
        if not isinstance(command, AsciiCommand):
            raise TypeError("write must be passed a string or AsciiCommand.")
        logger.debug("> %s", command)

        # From "Porting Python 2 Code to Python 3":
        # "...when you receive text in binary data, you should
        # immediately decode it. And if your code needs to send text as
        # binary data then encode it as late as possible.
        # This allows your code to work with only [unicode] text
        # internally and thus eliminates having to keep track of what
        # type of data you are working with."
        # See https://docs.python.org/3/howto/pyporting.html#text-versu
        # s-binary-data
        self._ser.write(command.encode())

    def read(self):
        """Reads a reply from the serial port.

        Raises:
            zaber.serial.TimeoutError: The duration specified by *timeout* 
                elapsed before a full reply could be read.
            ValueError: The reply read could not be parsed and is
                invalid.

        Returns:
            An `AsciiReply` containing the reply received.
        """
        line = self._ser.readline()
        if not line:
            logger.debug("< Receive timeout!")
            raise TimeoutError("read timed out.")
        decoded_line = line.decode()
        logger.debug("< %s", decoded_line.rstrip("\r\n"))
        return AsciiReply(decoded_line)

    def flush(self):
        """Flushes the buffers of the underlying serial port."""
        self._ser.flush()

    def open(self):
        """Opens the serial port."""
        self._ser.open()

    def close(self):
        """Closes the serial port."""
        self._ser.close()

    def __enter__(self):
        return self 

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def timeout(self):
        return self._ser.timeout

    @timeout.setter
    def timeout(self, t):
        self._ser.timeout = t

    @property
    def baudrate(self):
        return self._ser.baudrate

    @baudrate.setter
    def baudrate(self, b):
        if b not in (115200, 57600, 38400, 19200, 9600):
            raise ValueError("Invalid baud rate: {:d}. Valid baud rates are "
                    "115200, 57600, 38400, 19200, and 9600.".format(b))
        self._ser.baudrate = b




import pytest

from zaber.serial import BinaryCommand

def test_3_argument_constructor():
    bc = BinaryCommand(1, 5, 334)
    assert(bc.device_number == 1)
    assert(bc.command_number == 5)
    assert(bc.data == 334)

def test_message_id_in_4th_arg():
    bc = BinaryCommand(1, 2, 3, 4)
    assert(bc.device_number == 1)
    assert(bc.command_number == 2)
    assert(bc.data == 3)
    assert(bc.message_id == 4)

def test_message_id_is_none():
    bc = BinaryCommand(2, 54, 34)
    assert(bc.device_number == 2)
    assert(bc.command_number == 54)
    assert(bc.data == 34)
    assert(bc.message_id == None)

def test_2_argument_constructor():
    bc = BinaryCommand(1, 0)
    assert(bc.device_number == 1)
    assert(bc.command_number == 0)
    assert(bc.data == 0)
    assert(bc.message_id == None)

def test_no_arguments_throws_typeerror():
    with pytest.raises(TypeError):
        BinaryCommand()

def test_negative_device_or_command_number_raises_valueerror():
    with pytest.raises(ValueError):
        BinaryCommand(-1, 1, 0)
    with pytest.raises(ValueError):
        BinaryCommand(3, -2, 123, 3)

def test_negative_data_does_not_raise_exceptions():
    BinaryCommand(1, 55, -12312)

def test_encode():
    bc = BinaryCommand(1, 2, 3)
    # In this case encode() is not the inverse of decode().
    # encode() will convert from a BinaryCommand to "bytes",
    # and then decode will convert from bytes to unicode str.
    # The "decode()" call is only necessary to make this test
    # consistent in both Python 2 and 3.
    s = bc.encode().decode()
    assert(s[0] == '\x01')
    assert(s[1] == '\x02')
    assert(s[2] == '\x03')
    assert(s[3] == s[4] == s[5] == '\x00')

def test_encode_with_message_id():
    bc = BinaryCommand(1, 2, 3, 4)
    s = bc.encode().decode()
    assert(s[0] == '\x01')
    assert(s[1] == '\x02')
    assert(s[2] == '\x03')
    assert(s[3] == s[4] == '\x00')
    assert(s[5] == '\x04')

def test_message_id_clobbers_4th_data_bit():
    bc = BinaryCommand(1, 2, 2038004089, 9)
    s = bc.encode().decode()
    assert(s[0] == '\x01')
    assert(s[1] == '\x02')
    assert(s[2] == s[3] == s[4] == '\x79')
    assert(s[5] == '\x09')



"""The 'binary' module contains all classes related to the Binary 
protocol.
"""

import serial
import struct
import logging

from .exceptions import TimeoutError, UnexpectedReplyError

# See https://docs.python.org/2/howto/logging.html#configuring-logging-
# for-a-library for info on why we have these two lines here.
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class BinaryCommand(object):
    """Models a single command in Zaber's Binary protocol.

    Attributes:
        device_number: An integer representing the number (*a.k.a.*
            address) of the device to which to send the command. A 
            device number of 0 indicates the command should be executed
            by all devices. 0-255.
        command_number: An integer representing the command to be sent
            to the device. Command numbers are listed in Zaber's
            `Binary Protocol Manual`_. 0-255.
        data: The data value to be transmitted with the command.
        message_id: The `message ID`_ of the command. 0-255, or None if
            not present.

    .. _Binary Protocol Manual: http://www.zaber.com/wiki/Manuals/Binary
        _Protocol_Manual#Quick_Command_Reference
    .. _message ID: http://www.zaber.com/wiki/Manuals/Binary_Protocol_Ma
        nual#Set_Message_Id_Mode_-_Cmd_102
    """
    def __init__(self, device_number, command_number, data = 0, 
            message_id = None):
        """
        Args:
            device_number: An integer specifying the number of the
                target device to which to send this command. 0-255.
            command_number: An integer specifying the command to be
                sent. 0-255.
            data: An optional integer containing the data value to be
                sent with the command. When omitted, *data* will be set
                to 0.
            message_id: An optional integer specifying a message ID to
                give to the message. 0-255, or None if no message ID is
                to be used.

        Raises:
            ValueError: An invalid value was passed.
        """
        if device_number < 0 or command_number < 0:
            raise ValueError("Device and command number must be between 0 "
                    "and 255.")
        self.device_number = device_number
        self.command_number = command_number
        self.data = data
        if message_id is not None and (message_id < 0 or message_id > 255):
            raise ValueError("Message ID must be between 0 and 255.")
        self.message_id = message_id

    def encode(self):
        """Encodes a 6-byte byte string to be transmitted to a device.

        Returns:
            A byte string of length 6, formatted according to Zaber's
            `Binary Protocol Manual`_.
        """
        packed = struct.pack("<2Bl",
                self.device_number, self.command_number, self.data)
        if self.message_id is not None:
            packed = packed[:5] + struct.pack("B", self.message_id)
        return packed

    def __str__(self):
        return "[{:d}, {:d}, {:d}]".format(self.device_number,
                self.command_number, self.data)

class BinaryDevice(object):
    """A class to represent a Zaber device in the Binary protocol.

    Attributes:
        port: A BinarySerial object which represents the port to which
            this device is connected.
        number: The integer number of this device. 1-255.
    """
    def __init__(self, port, number):
        """
        Args:
            port: A BinarySerial object to use as a parent port.
            number: An integer between 1 and 255 which is the number of
                this device.

        Raises:
            ValueError: The device number was invalid.
        """
        if number > 255 or number < 1:
            raise ValueError("Device number must be 1-255.")
        self.number = number
        self.port = port

    def send(self, *args):
        """Sends a command to this device, then waits for a response.

        Args:
            *args: Either a single BinaryCommand, or 1-3 integers
                specifying, in order, the command number, data value,
                and message ID of the command to be sent.

        Notes:
            The ability to pass integers to this function is provided
            as a convenience to the programmer. Calling 
            ``device.send(2)`` is equivalent to calling 
            ``device.send(BinaryCommand(device.number, 2))``.

            Note that in the Binary protocol, devices will only reply
            once they have completed a command. Since this function
            waits for a reply from the device, this function may block
            for a long time while it waits for a response. For the same
            reason, it is important to set the timeout of this device's
            parent port to a value sufficiently high that any command
            sent will be completed within the timeout.

            Regardless of the device address specified to this function,
            the device number of the transmitted command will be 
            overwritten with the number of this device.

            If the command has a message ID set, this function will return
            a reply with a message ID. It does not check whether the message
            IDs match.

        Raises:
            UnexpectedReplyError: The reply read was not send by this
                device.

        Returns: A BinaryReply containing the reply received.
        """
        if len(args) == 1 and isinstance(args[0], BinaryCommand):
            command = args[0]
        elif len(args) < 4:
            command = BinaryCommand(self.number, *args)

        command.device_number = self.number
        self.port.write(command)
        reply = self.port.read(command.message_id is not None)

        if reply.device_number != self.number:
            raise UnexpectedReplyError("Received an unexpected reply from "
                    "device number {0:d}".format(reply.device_number),
                    reply)
        return reply

    def home(self):
        """Sends the "home" command (1), then waits for the device to
        reply.
        
        Returns: A BinaryReply containing the reply received.
        """
        return self.send(1)

    def move_abs(self, position):
        """Sends the "move absolute" command (20), then waits for the
        device to reply.

        Args:
            position: The position in microsteps to which to move.
        
        Returns: A BinaryReply containing the reply received.
        """
        return self.send(20, position)

    def move_rel(self, distance):
        """Sends the "move relative" command (21), then waits for the
        device to reply.
        
        Args:
            distance: The distance in microsteps to which to move.
        
        Returns: A BinaryReply containing the reply received.
        """
        return self.send(21, distance)

    def move_vel(self, speed):
        """Sends the "move at constant speed" command (22), then waits
        for the device to reply.

        Args:
            speed: An integer representing the speed at which to move.

        Notes:
            Unlike the other "move" commands, the device replies
            immediately to this command. This means that when this
            function returns, it is likely that the device is still
            moving.
        
        Returns: A BinaryReply containing the reply received.
        """
        return self.send(22, speed)

    def stop(self):
        """Sends the "stop" command (23), then waits for the device to
        reply.
        
        Returns: A BinaryReply containing the reply received.
        """
        return self.send(23)

    def get_status(self):
        """Sends the "Return Status" command (54), and returns the 
        result.

        Returns:
            An integer representing a `status code`_, according to
            Zaber's Binary Protocol Manual.

        .. _status code: http://www.zaber.com/wiki/Manuals/Binary_Protoc
            ol_Manual#Return_Status_-_Cmd_54
        """
        return self.send(54).data



class BinaryReply(object):
    """Models a single reply in Zaber's Binary protocol.

    Attributes:
        device_number: The number of the device from which this reply
            was sent.
        command_number: The number of the command which triggered this
            reply.
        data: The data value associated with the reply.
        message_id: The message ID number, if present, otherwise None.
    """
    def __init__(self, reply, message_id = False):
        """
        Args:
            reply: A byte string of length 6 containing a binary reply
                encoded according to Zaber's Binary Protocol Manual.
            message_id: True if a message ID should be extracted from
                the reply, False if not.

        Notes:
            Because a Binary reply's message ID truncates the last byte
            of the data value of the reply, it is impossible to tell
            whether a reply contains a message ID or not. Therefore, the
            user must specify whether or not a message ID should be
            assumed to be present.

        Raises:
            TypeError: An invalid type was passed as *reply*. This may
                indicate that a unicode string was passed instead of a
                binary (ascii) string.
        """
        if isinstance(reply, bytes):
            self.device_number, self.command_number, self.data = \
                    struct.unpack("<2Bl", reply)
            if (message_id):
                # Use bitmasks to extract the message ID.
                self.message_id = (self.data & 0xFF000000) >> 24
                self.data = self.data & 0x00FFFFFF 
  
                # Sign extend 24 to 32 bits in the message ID case.
                # If the data is more than 24 bits it will still be wrong,
                # but now negative smaller values will be right.
                if 0 != (self.data & 0x00800000):
                    self.data = (int)((self.data | 0xFF000000) - (1 << 32))
            else: 
                self.message_id = None

        elif isinstance(reply, list):
            # Assume a 4th element is a message ID.
            if len(reply) > 3: message_id = True
            self.device_number = reply[0]
            self.command_number = reply[1]
            self.data = reply[2]
            self.message_id = reply[3] if message_id else None

        else:
            raise TypeError("BinaryReply must be passed a byte string "
                    "('bytes' type) or a list.")

    def encode(self):
        """Returns the reply as a binary string, in the form in which it
        would appear if it had been read from the serial port.

        Returns:
            A byte string of length 6 formatted according to the Binary
            Protocol Manual.
        """
        return struct.pack("<2Bl", self.device_number,
                self.command_number, self.data)

    def __str__(self):
        return "[{:d}, {:d}, {:d}]".format(self.device_number, 
                self.command_number, self.data)

class BinarySerial(object):
    """A class for interacting with Zaber devices using the Binary protocol.

    This class defines a few simple methods for writing to and reading
    from a device connected over the serial port.
    """

    def __init__(self, port, baud = 9600, timeout = 5, inter_char_timeout = 0.01):
        """Creates a new instance of the BinarySerial class.

        Args:
            port: A string containing the name of the serial port to
                which to connect.
            baud: An integer representing the baud rate at which to
                communicate over the serial port.
            timeout: A number representing the number of seconds to wait
                for a reply. Fractional numbers are accepted and can be
                used to specify times shorter than a second.
            inter_char_timeout : A number representing the number of seconds
                to wait between bytes in a reply. If your computer is bad at
                reading incoming serial data in a timely fashion, try 
                increasing this value.

        Notes:
            This class will open the port immediately upon
            instantiation. This follows the pattern set by PySerial,
            which this class uses internally to perform serial
            communication.

        Raises:
            TypeError: The port argument passed was not a string.
        """
        if not isinstance(port, str):
            raise TypeError("port must be a string.")
        try:
            self._ser = serial.serial_for_url(port, do_not_open=True)
            self._ser.baudrate = baud
            self._ser.timeout = timeout
            self._ser.interCharTimeout = inter_char_timeout
            self._ser.open()
        except AttributeError:
            # serial_for_url not supported; use fallback
            self._ser = serial.Serial(port, baud, timeout = timeout, interCharTimeout = inter_char_timeout)

    def write(self, *args):
        r"""Writes a command to the port.

        This function accepts either a BinaryCommand object, a set
        of integer arguments, a list of integers, or a string. 
        If passed integer arguments or a list of integers, those
        integers must be in the same order as would be passed to the
        BinaryCommand constructor (ie. device number, then command
        number, then data, and then an optional message ID).

        Args:
            *args: A BinaryCommand to be sent, or between 2 and 4
                integer arguements, or a list containing between 2 and
                4 integers, or a string representing a 
                properly-formatted Binary command.
                
        Notes:
            Passing integers or a list of integers is equivalent to
            passing a BinaryCommand with those integers as constructor
            arguments.

            For example, all of the following are equivalent::

                >>> write(BinaryCommand(1, 55, 1000))
                >>> write(1, 55, 1000)
                >>> write([1, 55, 1000])
                >>> write(struct.pack("<2Bl", 1, 55, 1000))
                >>> write('\x01\x37\xe8\x03\x00\x00')

        Raises:
            TypeError: The arguments passed to write() did not conform
                to the specification of ``*args`` above.
            ValueError: A string of length other than 6 was passed.
        """
        if len(args) == 1:
            message = args[0]
            if isinstance(message, list):
                message = BinaryCommand(*message)
        elif 1 < len(args) < 5:
            message = BinaryCommand(*args)
        else:
            raise TypeError("write() takes at least 1 and no more than 4 "
                    "arguments ({0:d} given)".format(len(args)))

        if isinstance(message, str):
            logger.debug("> %s", message)
            data = message
            if len(data) != 6:
                raise ValueError("write of a string expects length 6.")

        elif isinstance(message, BinaryCommand):
            data = message.encode()
            logger.debug("> %s", message)

        else:
            raise TypeError("write must be passed several integers, or a "
                    "string, list, or BinaryCommand.")

        self._ser.write(data)

    def read(self, message_id = False):
        """Reads six bytes from the port and returns a BinaryReply.

        Args:
            message_id: True if the response is expected to have a 
                message ID. Defaults to False.

        Returns:
            A BinaryCommand containing all of the information read from
            the serial port.

        Raises: 
            zaber.serial.TimeoutError: No data was read before the 
                specified timeout elapsed.
        """
        reply = self._ser.read(6)
        if len(reply) != 6:
            logger.debug("< Receive timeout!")
            raise TimeoutError("read timed out.")
        parsed_reply = BinaryReply(reply, message_id)
        logger.debug("< %s", parsed_reply)
        return parsed_reply

    def flush(self):
        """Flushes the buffers of the underlying serial port."""
        self._ser.flush()

    def open(self):
        """Opens the serial port."""
        self._ser.open()

    def close(self):
        """Closes the serial port."""
        self._ser.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._ser.close()

    @property
    def timeout(self):
        """The number of seconds to wait for input while reading.
        
        The ``timeout`` property accepts floating point numbers for
        fractional wait times.
        """
        return self._ser.timeout

    @timeout.setter
    def timeout(self, t):
        self._ser.timeout = t

    @property
    def baudrate(self):
        """The baud rate at which to read and write.

        The default baud rate for the Binary protocol is 9600. T-Series
        devices are only capable of communication at 9600 baud.
        A-Series devices can communicate at 115200, 57600, 38400,
        19200, and 9600 baud.

        Note that this changes the baud rate of the computer on which
        this code is running. It does not change the baud rate of
        connected devices.
        """
        return self._ser.baudrate

    @baudrate.setter
    def baudrate(self, b):
        if b not in (115200, 57600, 38400, 19200, 9600):
            raise ValueError("Invalid baud rate: {:d}. Valid baud rates are "
                    "115200, 57600, 38400, 19200, and 9600.".format(b))
        self._ser.baudrate = b




import pytest

from zaber.serial import AsciiReply

def test_reply_parsing():
    a = AsciiReply("@01 0 OK IDLE -- 0\r\n")
    assert(a.message_type == '@')
    assert(a.device_address == 1)
    assert(a.axis_number == 0)
    assert(a.reply_flag == "OK")
    assert(a.device_status == "IDLE")
    assert(a.warning_flag == "--")
    assert(a.data == "0")

def test_info_parsing():
    a = AsciiReply("#01 0 Visit www.zaber.com for instruction manuals.\r\n")
    assert(a.message_type == '#')
    assert(a.device_address == 1)
    assert(a.axis_number == 0)
    assert(a.reply_flag is None)
    assert(a.data == "Visit www.zaber.com for instruction manuals.")

def test_info_parsing_trac1292():
    a = AsciiReply("#01 0 47 ")
    assert(a.message_type == '#')
    assert(a.data == "")

def test_alert_parsing():
    a = AsciiReply("!01 0 IDLE --\r\n")
    assert(a.message_type == '!')
    assert(a.device_address == 1)
    assert(a.axis_number == 0)
    assert(a.device_status == "IDLE")
    assert(a.warning_flag == "--")
    assert(a.reply_flag is None)
 
def test_message_ids():
    a = AsciiReply("@04 2 12 OK IDLE -- 123\r\n")
    assert(a.message_id == 12)
    assert(a.message_type == '@')
    assert(a.device_address == 4)
    assert(a.axis_number == 2)
    assert(a.reply_flag == "OK")
    assert(a.device_status == "IDLE")
    assert(a.warning_flag == "--")
    assert(a.data == "123")
    
    b = AsciiReply("#03 0 13 Visit www.zaber.com for instruction manuals.\r\n")
    assert(b.message_type == '#')
    assert(b.device_address == 3)
    assert(b.axis_number == 0)
    assert(b.message_id == 13)
    assert(b.data == "Visit www.zaber.com for instruction manuals.")

def test_invalid_string_raises_exception():
    with pytest.raises(ValueError):
        AsciiReply("%33 A OK I AM A BANANA\r\n")
    with pytest.raises(ValueError):
        AsciiReply("4")
    with pytest.raises(ValueError):
        AsciiReply("#BD 4 !! \n")
    with pytest.raises(ValueError):
        AsciiReply("@01 0 OK IDLE --\r\n")
    with pytest.raises(ValueError):
        AsciiReply("@01 0 OK IDLE -- \r\n")

def test_encoding_strings():
    reply_string = "@01 0 OK IDLE -- 0\r\n"
    a = AsciiReply(reply_string)
    assert(a.encode() == reply_string)

def test_encode_with_message_id():
    repstr = "@01 0 12 OK IDLE -- 0\r\n"
    a = AsciiReply(repstr)
    assert(a.encode() == repstr)

def test_encode_with_checksum():
    repstr = "@01 0 OK IDLE -- 0:8D\r\n"
    a = AsciiReply(repstr)
    assert(a.encode() == repstr)

def test_encode_with_single_digit_checksum():
    repstr = "@01 0 OK IDLE -- 7z:0C\r\n"
    a = AsciiReply(repstr)
    assert(a.encode() == repstr)

def test_bad_checksum_raises_exception():
    repstr = "@01 0 OK IDLE -- 0:GG\r\n"
    with pytest.raises(ValueError):
        a = AsciiReply(repstr)



import os
import tty
import time
import select
import re
import struct

from zaber.serial import BinaryCommand

class FakePort(object):
    def __init__(self):
        self._master, self.slave = os.openpty()
        tty.setraw(self._master)
        tty.setraw(self.slave)

    def expect(self, pattern, timeout = 5):
        """Block until a match is read from the master pty.

        This function will read in bytes one-at-a-time, append them to
        a buffer, then concatenate the buffer's elements into a string.
        That string will be checked against the pattern provided, and
        if a match is found then the function returns.

        By default, this function has a five-second timeout. If None is
        passed as the timeout argument, then the function will block
        forever until the expected input is read.

        Args:
            pattern: A string containing a regular expression which
                describes the expected input to be read.
            timeout: An integer representing the number of seconds to
                wait before giving up on the match. This wait is an
                inter-byte wait, not a total wait.
        """
        if isinstance(pattern, bytes):
            regex = re.compile(pattern)
        else: regex = re.compile(pattern.encode())
        buffer = bytearray()
        
        # If None is passed for timeout, omit the timeout argument to select.
        # From the docs: "When the timeout argument is omitted the function 
        # blocks until at least one file descriptor is ready."
        selectargs = ([self._master], [], [], timeout) if timeout is not None \
                else ([self._master], [], [])

        while 1:
            r, w, x = select.select(*selectargs)
            if r == []:
                print('Pattern "{0!s}" not found. Read "{1!s}".'.format(
                    pattern, bytes(buffer)))
                raise TimeoutError("Select timed out.")

            c = os.read(self._master, 1)
            buffer.append(ord(c))

            if regex.search(bytes(buffer)):
                break

            time.sleep(0.01)

    def send(self, *args):
        """Send a message from the master pty to the slave.

        Args:
            *args: Between 1 and 3 arguments. If there is only one
                argument, then the string representation (ie. the value
                returned from the object's __str__ function) of the
                argument is written to the serial port.

                If there are two or three arguments, they are assumed
                to be the device number, command number, and data value
                of a Binary command according to Zaber's Binary
                Protocol Manual.
        """
        if len(args) == 0:
            raise TypeError("send() takes at least 1 argument (0 given)")

        # Write whatever is given as a string.
        if len(args) == 1:
            if isinstance(args[0], bytes):
                os.write(self._master, args[0])
            else: os.write(self._master, args[0].encode())

        # Assume 2 or 3 arguments are binary replies and pack them.
        if len(args) == 2:
            os.write(self._master, struct.pack("<2Bl", args[0], args[1], 0))

        if len(args) == 3:
            os.write(self._master, struct.pack("<2Bl", *args))

        if len(args) > 3:
            raise TypeError("send() takes at most 3 arguments")

class TimeoutError(Exception):
    pass




import pytest
import struct

from zaber.serial import BinaryCommand, BinarySerial

def test_write(binaryserial, fake):
    cmd = BinaryCommand(1, 54)
    binaryserial.write(cmd)
    fake.expect(cmd.encode())

def test_read(binaryserial, fake):
    fake.send(1, 54, 0)
    rep = binaryserial.read().encode().decode()
    assert(rep[0] == '\x01')
    assert(rep[1] == '\x36')
    assert(rep[2] == rep[3] == rep[4] == rep[5] == '\x00')

def test_write_multiple_arguments(binaryserial, fake):
    binaryserial.write(1, 2, 3)
    fake.expect(pack(1, 2, 3))

def test_write_multi_args_with_message_id(binaryserial, fake):
    binaryserial.write(1, 2, 3, 4)
    fake.expect(pack(1, 2, 3, 4))

def test_write_string_trac1276(binaryserial, fake):
    binaryserial.write("\x01\x02\x03\x04\x05\x06")
    fake.expect(pack(1, 2, 3 | 4 << 8 | 5 << 16 | 6 << 24))

def test_write_string_wrong_length_trac1276(binaryserial, fake):
    with pytest.raises(ValueError):
        binaryserial.write("\x01\x02\x03\x04\x05\x06\x07")

def test_write_complains_when_passed_wrong_type(binaryserial):
    with pytest.raises(TypeError):
        binaryserial.write({"aaaa": "bbb"})
    with pytest.raises(TypeError):
        binaryserial.write(binaryserial)

def test_constructor_fails_when_not_passed_a_string():
    with pytest.raises(TypeError):
        BinarySerial(1)

def test_read_reads_message_id(binaryserial, fake):
    fake.send(BinaryCommand(1, 55, 34, 22))
    reply = binaryserial.read(message_id = True)
    assert(reply.device_number == 1)
    assert(reply.command_number == 55)
    assert(reply.data == 34)
    assert(reply.message_id == 22)

def test_read_truncates_data_when_using_message_id(binaryserial, fake):
    fake.send(BinaryCommand(3, 55, 0x117FFFFF, 1)) # the 7 is to avoid sign extension.
    reply = binaryserial.read(True)
    assert(reply.device_number == 3)
    assert(reply.command_number == 55)
    assert(reply.data == 0x007FFFFF)
    assert(reply.message_id == 1)

def pack(device, command, data = 0, message_id = None):
    packed = struct.pack("<2Bl", device, command, data)
    if message_id is not None:
        packed = packed[:5] + struct.pack("B", message_id)
    return packed

def test_port_not_string_trac1365():
    with pytest.raises(TypeError):
        ser = BinarySerial(1)

def test_port_url_works_trac1362():
    ser = BinarySerial("loop://")



import pytest
import threading
import time

from zaber.serial import AsciiSerial, AsciiCommand, AsciiReply, BinaryCommand, TimeoutError

def test_write_command(asciiserial, fake):
    a = AsciiCommand("/1 echo TEST\r\n")
    asciiserial.write(a)
    asciiserial.write("/1 echo TEST\r\n")
    asciiserial.write("1 echo TEST")
    fake.expect("/1 0 echo TEST\r\n")
    fake.expect("/1 0 echo TEST\r\n")
    fake.expect("/1 0 echo TEST\r\n")

def test_write_empty_string(asciiserial, fake):
    asciiserial.write("")
    fake.expect("/0 0\s*\r\n")
    asciiserial.write("1")
    fake.expect("/1 0\r\n")
    asciiserial.write("2 2")
    fake.expect("/2 2\r\n")

def test_write_string(asciiserial, fake):
    asciiserial.write("home")
    fake.expect("/0 0 home\r\n")
    asciiserial.write("1 move abs 2000")
    fake.expect("/1 0 move abs 2000\r\n")
    asciiserial.write("5 tools setcomm 9600 1")
    fake.expect("/5 0 tools setcomm 9600 1\r\n")
    asciiserial.write("3 2 move rel -200")
    fake.expect("/3 2 move rel -200\r\n")

def test_write_binary_bytes(asciiserial, fake):
    asciiserial.write(b"1 home\r\n")
    fake.expect("/1 0 home\r\n")

def test_read_returns_command(asciiserial, fake):
    fake.send("@01 0 OK IDLE -- 0\r\n")
    reply = asciiserial.read()
    assert(isinstance(reply, AsciiReply))
    assert(reply.device_address == 1)
    assert(reply.axis_number == 0)
    assert(reply.reply_flag == 'OK')
    assert(reply.device_status == 'IDLE')
    assert(reply.warning_flag == '--')
    assert(reply.data == '0')

def test_read_complains_on_malformed_input(asciiserial, fake):
    fake.send("bad input\r\n")
    with pytest.raises(ValueError):
        asciiserial.read()

def test_can_set_timeout(asciiserial, fake):
    asciiserial.timeout += 10
    t = threading.Thread(target = asciiserial.read)
    t.start()

    # If the timeout hasn't been properly set, 
    # this sleep will trigger a TimeoutError.
    time.sleep(asciiserial.timeout - 9)
    fake.send("@01 0 OK IDLE -- 0\r\n")
    t.join()
    asciiserial.timeout -= 10 

def test_can_change_baudrate(asciiserial):
    # If _ser gets renamed, then this test will need to be changed too.
    underlying_serial = asciiserial._ser
    for rate in (9600, 19200, 38400, 57600, 115200):
        asciiserial.baudrate = rate 
        assert(underlying_serial.baudrate == rate)

def test_sending_BinaryCommand_raises_error(asciiserial):
    with pytest.raises(TypeError):
        asciiserial.write(BinaryCommand(1, 55, 23423))

def test_read_times_out_with_empty_buffer(asciiserial):
    with pytest.raises(TimeoutError):
        asciiserial.read()

def test_close_and_reopen(asciiserial):
    asciiserial.close()
    asciiserial.open()

def test_port_not_string_trac1365():
    with pytest.raises(TypeError):
        ser = AsciiSerial(1)

def test_port_url_works_trac1362():
    ser = AsciiSerial("loop://")



import pytest
from threading import Thread

from zaber.serial import AsciiAxis, AsciiDevice

@pytest.fixture
def device(asciiserial):
    return AsciiDevice(asciiserial, 1)

@pytest.fixture
def axis(device):
    return AsciiAxis(device, 1)

def test_constructor(device):
    a = AsciiAxis(device, 1)
    assert(a.number == 1)
    assert(a.parent is not None)

def test_can_be_constructed_from_asciidevice(device):
    a = device.axis(1)
    b = device.axis(2)
    assert(a.number == 1)
    assert(b.number == 2)

def test_home(axis, fake):
    t = Thread(target = axis.home)
    t.start()
    fake.expect("/1 1 home\r\n")
    fake.send("@01 1 OK BUSY WR 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    t.join()

def test_move_abs(axis,fake):
    t = Thread(target = axis.move_abs, args = (10234, ))
    t.start()
    fake.expect("/1 1 move abs 10234\r\n")
    fake.send("@01 1 OK BUSY -- 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    t.join()

def test_poll_until_idle_will_continue_polling(axis, fake):
    t = Thread(target = axis.poll_until_idle)
    t.start()
    for i in range(100):
        fake.expect("/1 1\r\n")
        fake.send("@01 1 OK BUSY -- 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    t.join()

def test_move_rel(axis, fake):
    t = Thread(target = axis.move_rel, args = (12312, ))
    t.start()
    fake.expect("/1 1 move rel 12312\r\n")
    fake.send("@01 1 OK BUSY -- 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    t.join()

def test_move_vel(axis, fake):
    t = Thread(target = axis.move_vel, args = (1005, ))
    t.start()
    fake.expect("/1 1 move vel 1005\r\n")
    fake.send("@01 1 OK BUSY -- 0\r\n")
    t.join()

def test_stop(axis, fake):
    t = Thread(target = axis.stop)
    t.start()
    fake.expect("/1 1 stop\r\n")
    fake.send("@01 1 OK BUSY NI 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    t.join()

def test_invalid_axis_number_raises_value_error(device):
    with pytest.raises(ValueError):
        AsciiAxis(device, 10)

def test_get_status(axis, fake):
    def test_return_value_from_get_status():
        assert(axis.get_status() == "IDLE")
        assert(axis.get_status() == "BUSY")
    t = Thread(target = test_return_value_from_get_status)
    t.start()
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK IDLE -- 0\r\n")
    fake.expect("/1 1\r\n")
    fake.send("@01 1 OK BUSY -- 0\r\n")
    t.join()



from .ascii import AsciiAxis, AsciiDevice, AsciiCommand, AsciiReply, AsciiSerial
from .binary import BinaryCommand, BinaryDevice, BinaryReply, BinarySerial
from .exceptions import TimeoutError, UnexpectedReplyError



from zaber.serial import BinaryCommand, BinaryDevice, BinaryReply, UnexpectedReplyError

import pytest, struct, threading

@pytest.fixture(scope="module")
def device(binaryserial):
    return BinaryDevice(binaryserial, 1)

def test_constructor(binaryserial):
    bd1 = BinaryDevice(binaryserial, 1)
    bd2 = BinaryDevice(binaryserial, 2)
    assert(bd1.number == 1)
    assert(bd2.number == 2)
    assert(bd1.port == bd2.port == binaryserial)

def test_send(device, fake):
    t = threading.Thread(target = device.send, args = (1, 2, 3))
    t.start()
    fake.expect(pack(device.number, 1, 2, 3))
    fake.send(pack(device.number, 1, 44, 3))
    t.join()

def test_send_overwrites_device_number(device, fake):
    t = threading.Thread(target = device.send,
            args = (BinaryCommand(device.number + 1, 1, 2), ))
    t.start()
    fake.expect(pack(device.number, 1, 2))
    fake.send(pack(device.number, 1, 0))
    t.join()

def test_send_complains_on_unexpected_response(device, fake):
    def device_send_with_pytest_raises():
        with pytest.raises(UnexpectedReplyError):
            device.send(2, 3, 4)
    t = threading.Thread(target = device_send_with_pytest_raises)
    t.start()
    fake.expect(pack(device.number, 2, 3, 4))
    fake.send(pack(device.number + 1, 2, 3, 4))
    t.join()

def test_send_complains_if_message_id_does_not_match(device, fake):
    def device_send_with_pytest_raises():
        with pytest.raises(UnexpectedReplyError):
            device.send(3, 3, 3)
    t = threading.Thread(target = device_send_with_pytest_raises)
    t.start()
    fake.expect(pack(device.number, 3, 3, 3))
    fake.send(pack(device.number, 3, 3, 45))
    t.join()

def test_send_with_id_receives_with_id(device, fake):
    def device_send(results):
        results.append(device.send(3, 3, 7))
    replies = []
    t = threading.Thread(target = device_send, args=(replies,))
    t.start()
    fake.expect(pack(device.number, 3, 3, 7))
    fake.send(pack(device.number, 3, 3, 7))
    t.join()
    assert(len(replies) == 1)
    assert(replies[0].message_id == 7)

def test_send_without_id_receives_without_id(device, fake):
    def device_send(results):
        results.append(device.send(3, 3))
    replies = []
    t = threading.Thread(target = device_send, args=(replies,))
    t.start()
    fake.expect(pack(device.number, 3, 3))
    fake.send(pack(device.number, 3, 3))
    t.join()
    assert(len(replies) == 1)
    assert(replies[0].message_id is None)

def test_home(device, fake):
    t = threading.Thread(target = device.home)
    t.start()
    fake.expect(pack(device.number, 1, 0))
    fake.send(pack(device.number, 1, 0))
    t.join()

def test_moveabs(device, fake):
    t = threading.Thread(target = device.move_abs, args = (1000, ))
    t.start()
    fake.expect(pack(device.number, 20, 1000))
    fake.send(pack(device.number, 20 ,1000))
    t.join()

def test_moverel(device, fake):
    t = threading.Thread(target = device.move_rel, args = (1000, ))
    t.start()
    fake.expect(pack(device.number, 21, 1000))
    fake.send(pack(device.number, 21, 2000))
    t.join()

def test_movevel(device, fake):
    t = threading.Thread(target = device.move_vel, args = (200, ))
    t.start()
    fake.expect(pack(device.number, 22, 200))
    fake.send(pack(device.number, 22, 200))
    t.join()

def test_stop(device, fake):
    t = threading.Thread(target = device.stop)
    t.start()
    fake.expect(pack(device.number, 23, 0))
    fake.send(pack(device.number, 23, 23134))
    t.join()

def test_getstatus(device, fake):
    def control_fake():
        fake.expect(pack(device.number, 54, 0))
        fake.send(pack(device.number, 54, 0))
    t = threading.Thread(target = control_fake)
    t.start()
    status = device.get_status()
    t.join()
    assert(status == 0)

def pack(device, command, data = 0, message_id = None):
    packed = struct.pack("<2Bl", device, command, data)
    if message_id is not None:
        packed = packed[:5] + struct.pack("B", message_id)
    return packed




import pytest

from zaber.serial import BinaryReply

def test_init():
    br = BinaryReply([1, 56, 10502])
    assert(br.device_number == 1)
    assert(br.command_number == 56)
    assert(br.data == 10502)

def test_init_with_message_id():
    br = BinaryReply([2, 44, 21231, 12])
    assert(br.device_number == 2)
    assert(br.command_number == 44)
    assert(br.data == 21231)
    assert(br.message_id == 12)

def test_message_id_is_None():
    br = BinaryReply([3, 32, 0])
    assert(br.message_id == None)

def test_parsing():
    br = BinaryReply(b"\x01\xFF\x03\x00\x00\x00")
    assert(br.device_number == 1)
    assert(br.command_number == 0xFF == 255)
    assert(br.data == 3)

def test_parsing_with_message_id():
    br = BinaryReply(b"\x01\x02\x03\x04\x05\x06", message_id=True)
    assert(br.device_number == 1)
    assert(br.command_number == 2)
    assert(br.data == 3 | (4 << 8) | (5 << 16) == 328707)
    assert(br.message_id == 6)

def test_parsing_sign_extension_with_message_id():
    br = BinaryReply(b"\x01\x02\x03\x04\x85\x06", message_id=True)
    assert(br.device_number == 1)
    assert(br.command_number == 2)
    assert(br.data == 0xFF850403 - (1 << 32) == -8059901)
    assert(br.message_id == 6)

def test_parsing_without_message_id():
    br = BinaryReply(b"\x01\x02\x03\x04\x05\x06", message_id=False)
    assert(br.device_number == 1)
    assert(br.command_number == 2)
    assert(br.data == 100992003 == 3 | (4 << 8) | (5 << 16) | (6 << 24) )
    assert(br.message_id == None)

def test_dict_will_raise_typeerror():
    br = BinaryReply([1, 2, 3])
    br = BinaryReply(b"\x01\x02\x03\x04\x05\x06")
    with pytest.raises(TypeError):
        br = BinaryReply({
            'device number': 1,
            'command number': 2,
            'data': 3}) # This should fail.





"""The exceptions module contains special exceptions to be raised by
this library.
"""

class TimeoutError(Exception):
    """Raised when a read operation exceeded its specified time limit."""
    pass

class UnexpectedReplyError(Exception):
    """Raised when a reply was read from an unexpected device."""
    
    def __init__(self, message, reply = None):
        """
        Args:
            message: The error message to display.
            reply: The unexpected reply which caused this exception to
                be raised.

        Notes:
            This exception preserves the unexpected reply for which it
            was thrown. You can access it through the *reply* attribute.
        """
        super(UnexpectedReplyError, self).__init__(message)
        
        self.reply = reply




# A special file for pytest fixtures. 
# All fixtures defined here are accessible to
# all test_*.py files in this directory.

import pytest, os

from deviceemulator import FakePort
from zaber.serial import AsciiSerial, BinarySerial

# We need to create fakeport here first so that 
# the slave pty can be given to the *serial fixtures.
fakeport = FakePort()

@pytest.yield_fixture(scope="module")
def asciiserial():
    with AsciiSerial(os.ttyname(fakeport.slave)) as s:
        yield s

@pytest.yield_fixture(scope="module")
def binaryserial():
    with BinarySerial(os.ttyname(fakeport.slave)) as s:
        yield s

@pytest.fixture(scope="module")
def fake():
    return fakeport



from setuptools import setup, find_packages

from os import path

def read_from_file(*filename):
    with open(path.join(*filename), 'r') as f:
        return f.read()

setup(
    name = 'zaber.serial',
    version = '0.8',
    packages = find_packages(exclude=["test*", "test_*", "*_test*"]),
    description = 'A library for communicating with Zaber devices',
    long_description = read_from_file('DESCRIPTION.rst'),
    url = 'http://www.zaber.com/support/docs/api/core-python/',
    author = 'Zaber Technologies Inc.',
    author_email = 'contact@zaber.com',
    license = 'Apache Software License v2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4'
    ],
    keywords = 'serial zaber',

    # This library is built on top of PySerial.
    install_requires=['pyserial'],
    extras_require = { 'test': ['pytest'] }
)



import pytest
import threading

from zaber.serial import UnexpectedReplyError, TimeoutError, AsciiReply, BinaryReply, AsciiCommand, AsciiDevice

def test_unexpectedreplyerror_accepts_reply():
    rep = AsciiReply("@01 0 OK IDLE -- 0\r\n")
    ex = UnexpectedReplyError("Oh no!", rep)
    assert(ex.args == ("Oh no!", ))
    assert(ex.reply == rep)

def test_message_available_from_raises():
    try:
        raise UnexpectedReplyError("TEST", BinaryReply([1, 2, 3]))
    except UnexpectedReplyError as ex:
        assert(ex.args)
        assert(ex.reply)
        
def test_unexpectedreplyerror_contains_unexpected_reply(fake, asciiserial):
    def test_write_and_check_exception():
        cmd = AsciiCommand("1 1 move abs 200")
        device = AsciiDevice(asciiserial, 1)
        rep = None
        try:
            device.send(cmd)
        except UnexpectedReplyError as ure:
            rep = ure.reply
        assert(rep != None)
    t = threading.Thread(target = test_write_and_check_exception)
    t.start()
    fake.expect("/1 1 move abs 200\r\n")
    fake.send("@02 2 OK IDLE -- 0\r\n")
    t.join()
